#!/usr/bin/perl

#/////////////////////////////////////////////////////////////////////////////
use constant { ROLE => 11, NAME => 12, DESCRIPTION => 13, NOT_IN_TREE => 14, ROLE_NONE => 15, ROLE_PRESENTATION => 16 };
use constant { TABLE => "wikitable", COMMENT => "comment", TEST => "statement" };
use constant { COUNT => "count", JSON => "json", LABEL => "label", KEY => "key", ROWS => "rows", ROWSPAN => "ROWSPAN" };

=head2 populate_test_template
 
 Parameters  : template, test_description, testable_statement
 Returns     : string
 Description : Populates the template with the test description and testable statement 
 			   and returns the completed test string.
 
=cut

sub populate_test_template ($$$) {
	my $template = $_[0];

	# test for needed namespaces
	$xlink = index( $_[2], 'xlink' ) > -1;
	$html  = index( $_[2], '<html:' ) > -1;

	# Only add namespaces when needed, otherise remove XLINK placeholder from the template
	if ($xlink) {
		if ($html) {
			$template =~
			  s/XLINK/xmlns:xlink="http:\/\/www.w3.org\/1999\/xlink" xmlns:html="http:\/\/www.w3.org\/1999\/html" /;
		}
		else {
			# only xlink namepsace needed
			$template =~ s/XLINK/xmlns:xlink="http:\/\/www.w3.org\/1999\/xlink" /;
		}
	}
	elsif ($html) {

		# html namespace needed
		$template =~ s/XLINK/xmlns:html="http:\/\/www.w3.org\/1999\/html" /;
	}
	else {
		# no additional namespaces needed remove placeholder
		$template =~ s/XLINK//;
	}

	$template =~ s/COMMENT/$_[1]/;
	$template =~ s/TESTABLE_STATEMENT/$_[2]/;

	return $template;
}

#/////////////////////////////////////////////////////////////////////////////

=head2 test_type
 
 Parameters  : wiki header line
 Returns     : test type 
 Description : pulls the test type by finding the word after Testable statements for.
 
=cut

sub test_type($) {
	if ( index( $_[0], 'not included in accessibility tree' ) > -1 ) {
		return 'not in accessibility tree';
	}
	my $ind = index( $_[0], 'for' );
	my $type = substr( $_[0], $ind + 3 );

	$ind = index( $type, "==" );
	if ( $ind > 0 ) {
		$type = substr( $type, 0, $ind );
	}
	$type =~ s/^\s+|\s+$//go;    # trim
	return $type;
}

#/////////////////////////////////////////////////////////////////////////////

=head2 build_rowspan_map
 
 Parameters  : map, wikitable
 Returns     :  wikitable string
 Description : builds JSON for multi rows and replaces the used substring with a key
 
=cut

sub build_rowspan_map($$) {
	my $str   = $_[1];
	my @spans = split_clean( 'rowspan', $_[1] );    # split of the wiki table into rowspan trees
	my $len   = scalar @spans;
	for ( my $i = $len - 1 ; $i > 0 ; $i-- ) {
		populate_rowspan_map( $_[0], $i, \@spans );
	}
	return @spans;
}

#/////////////////////////////////////////////////////////////////////////////

=head2 is_noise
 
 Parameters  : string
 Returns     : 1 if noise otherwise 0
 Description : checks the string to see if it is useful or noise - ie a wiki table artifact, whitespace, line return...
 
=cut

sub is_noise($) {
	my $str  = clean_for_output( $_[0] );
	my $leng = length($str);

	# if the cleaned string is a pipe (|) it is noise (a wiki table artifact) and not useful
	return ( $leng == 0 || ( $leng == 1 && index( $str, '|' ) == 0 ) );
}

#/////////////////////////////////////////////////////////////////////////////

=head2 split_clean
 
 Parameters  :  regex, string  
 Returns     : array of cleaned items
 Description : splits the string, removes wiki artifacts from the ends
 
=cut

sub split_clean($$) {
	my @rows = split( $_[0], $_[1] );    # split of wiki table markup
	if ( is_noise( $rows[0] ) ) {
		shift @rows;
	}
	my $len = scalar @rows;
	if ($len) {
		if ( is_noise( $rows[ $len - 1 ] ) ) {
			pop @rows;
		}
	}

	return @rows;
}

#/////////////////////////////////////////////////////////////////////////////

=head2 row_json
 
 Parameters  : string for wiki table row
 Returns     : JSON 
 Description : makes JSON for a wiki table row
 
=cut

sub row_json($) {
	my @columns = split_clean( '\|', $_[0] );
	my $j = scalar @columns;
	if ( $j == 0 ) {
		return;
	}

	if ( $j == 1 ) {
		return clean_quote( $columns[0] ) . ': "SHOWN"';
	}
	elsif ( $j == 2 ) {
		return clean_quote( $columns[0] ) . ":" . clean_quote( $columns[1] );
	}
	elsif ( $j == 3 ) {
		return
		    clean_quote( $columns[0] ) . ": { "
		  . clean_quote( $columns[1] ) . ":"
		  . clean_quote( $columns[2] ) . " }";
	}
	elsif ( $j == 4 ) {
		return
		    clean_quote( $columns[0] ) . ": { "
		  . clean_quote( $columns[1] ) . ": { "
		  . clean_quote( $columns[2] ) . ":"
		  . clean_quote( $columns[3] ) . "} }";
	}
	else {
		# This is untested code
		# TODO test this code  
		
		my $str   = "";
		my $close = "";
		for ( $i = 0 ; $i < $j - 2; $i++ ) {
			$str   .= clean_quote( $columns[$i] ) . ": { ";
			$close .= "} ";
		}
		$str .= clean_quote( $columns[$j - 2] ) . ":" . clean_quote( $columns[$j - 1] ) . $close;
		return $str;
	}

	return;
}

#/////////////////////////////////////////////////////////////////////////////

=head2 rowspan_json
 
 Parameters  : JSON string, num rows, ref row array
 Returns     : JSON string
 Description : adds to the JSON string
 
=cut

sub rowspan_json($$$) {
	my $json = $_[0];
	my $max  = $_[1];
	my @rows = @{ $_[2] };
	for ( my $i = 0 ; $i < $max ; $i++ ) {
		my $row     = shift @rows;
		my $rowjson = row_json($row);
		if ( $rowjson && $i > 0 ) {
			$json .= ",";
		}
		if ($rowjson) {
			$json .= $rowjson;
		}

	}

	return ( JSON => $json, ROWS => \@rows );
}

#/////////////////////////////////////////////////////////////////////////////

=head2 populate_rowspan_map
 
 Parameters  : map, index, reference to rowspan split array
 Returns     :  spantext with key instead of the text used to build the rowspan
 Description : builds JSON for multi rows and replaces the used substring with a key
 
=cut

sub populate_rowspan_map($$$) {

	my $href    = $_[0];
	my $index   = $_[1];         # index for the array of rowspans
	my $spanref = $_[2];
	my @spans   = @{ $_[2] };    # array of rowspan splits

	my $group = $spanref->[$index];    # string for the rowspan we are working on

	#pull off the rowspan value ($rowcount), $label label of object spanning rows
	my @rowspan  = split( '\|', $group, 3 );
	my $rowcount = clean_for_output( $rowspan[0] );
	my $label    = clean_for_output( $rowspan[1] );

	# JSON from the multiple rows
	my $json    = '"' . $label . '": {';
	my $jsonlen = length($json);
	my @rows    = split_clean( '\|-', $rowspan[2] );    # split of the wiki table into rows
	my $max     = scalar @rows;
	my $delta   = $max - $rowcount;
	if ( $max > $rowcount ) {
		$max = $rowcount;
	}

	my %hash = rowspan_json( $json, $max, \@rows );
	$json = $hash{JSON};
	@rows = @{ $hash{ROWS} };
	my $nojsonentry = length($json) == $jsonlen;

	my $ix = $index + 1;
	while ( $delta < 0 ) {
		$group = $spanref->[$ix];
		my $pipeindex = index( $group, "|" );
		my $nextkey = ( $pipeindex > 0 ) ? substr( $group, 0, $pipeindex ) : $group;
		my $rsobj   = $$href{$nextkey};
		my $width   = $rsobj->{COUNT};
		$delta += $width;
		if ( $delta < 0 && $pipeindex > 0 ) {

			# call rowspan_json
			my $debug2 = 1;
		}
		if ($nojsonentry) {
			$json .= $rsobj->{JSON};
			$nojsonentry = 0;
		}
		else {
			$json .= ', ' . $rsobj->{JSON};
		}

		my $newstring = ( $pipeindex > 0 ) ? substr( $group, $pipeindex ) : '';
		$spanref->[$ix] = $newstring;
		$ix++;
	}

	$json .= "}";

	my $value = ROWSPAN . $index;

	my %spanobj = (
		LABEL => $label,
		JSON  => $json,
		COUNT => $rowcount,
		KEY   => $value
	);

	my $str = $value;
	if ( ( scalar @rows ) > 0 ) {
		unshift( @rows, $value );
		$str = join( '|-', @rows );
	}

	$$href{$value} = \%spanobj;
	$spanref->[$index] = $str;
}

#/////////////////////////////////////////////////////////////////////////////

=head2 rows_json_no_rowspans
 
 Parameters  : string, indent
 Returns     : JSON for rows in string
 Description : string is a a part of a wiki table that does not include any rowspans, builds the JSON for the rows
 
=cut

sub rows_json_no_rowspans($$) {
	my @singlerows = split_clean( '\|-', $_[0] );
	my $str        = '';
	my $len        = scalar @singlerows;
	my $needscomma = 0;
	for ( my $j = 0 ; $j < $len ; $j++ ) {
		if ( !is_noise( $singlerows[$j] ) ) {
			my $rjson = row_json( $singlerows[$j] );
			if ($rjson) {
				if ($needscomma) {
					$str .= ",\n";
				}

				if ($needsindent) {
					$str .= $_[1];
				}

				$str .= $rjson;

				$needscomma  = 1;
				$needsindent = 1;
			}
		}
	}
	return $str;
}

#/////////////////////////////////////////////////////////////////////////////

=head2 get_json_from_table
 
 Parameters  : wiki table (without {| |} lines), indent
 Returns     : JSON for expected result
 Description : builds JSON for the expected results wiki table
 
=cut

sub get_json_from_table($$) {

	# wiki table rowspans are handled first and the JSON is put in %spanmap
	my %spanmap;
	my @rows = build_rowspan_map( \%spanmap, $_[0] );

	# this function uses indents and line returns to make the JSON human readable
	my $str         = '';                    # JSON object string
	my $len         = scalar @rows;
	my $indent      = ' ' x ( $_[1] + 2 );
	my $needsindent = 0;

	for ( my $i = 0 ; $i < $len ; $i++ ) {
		my $row        = $rows[$i];                # the wiki table row
		my $rsind      = index( $row, ROWSPAN );
		my $needscomma = 0;
		if ( $rsind == -1 ) {
			my $rowstr = rows_json_no_rowspans( $row, $indent );
			if ( length($rowstr) > 0 ) {
				$str .= $rowstr;
				$needscomma  = 1;
				$needsindent = 1;
			}
		}
		else {
			my @spanrows = split( '\|-', $row, 2 );
			my $rowspanobj = $spanmap{ $spanrows[0] };
			if ($needsindent) {
				$str .= $indent;
			}
			if ( !$rowspanobj || !${$rowspanobj}{JSON} ) {
				my $debug1 = 1;
			}
			$str .= ${$rowspanobj}{JSON};
			$needscomma  = $i < $len - 1;
			$needsindent = 1;
			if ( scalar @spanrows > 1 ) {

				# add the rows not in the rowspan
				my $rs = rows_json_no_rowspans( $spanrows[1], $indent );
				if ($rs) {
					$str .= "," . $indent . $rs;
				} else {
					my $debug3 = 1;
				}
				
			}
		}

		if ($needscomma) {
			$str .= ",\n";    # add a comma and newline
		}

	}
	$str = "{ " . $str . "\n" . ' ' x $_[1] . "}\n";
	return $str;
}

#/////////////////////////////////////////////////////////////////////////////

=head2 read_wiki_file
 
 Parameters  : filename
 Returns     : hash of array of tests, key is test type (role, name, description)
 Description : gets the tests as a description [0] and a testable statement [1]
 
=cut

sub read_wiki_file ($) {
	open( FH, '<:encoding(UTF-8)', $_[0] ) or die "Could not open file '$_[0]' $!";
	my @lines = <FH>;
	close FH or die "Cannot close $_[0]: $!";

	my %alltests;
	my $tests;
	my $todos   = 0;
	my $counter = 0;
	my $commentline;
	my $testablestatement;
	my $max = @lines;
	for ( my $i = 0 ; $i < $max ; $i++ ) {
		if ( $lines[$i] =~ /^\s*==\s+/ && $lines[$i] =~ /\s+==\s*$/ ) {
			my $type     = test_type( $lines[$i] );
			my @testtype = ();
			$alltests{$type} = \@testtype;
			$tests           = \@testtype;
			$counter         = 0;
		}
		elsif ( $lines[$i] =~ /^;\s*\/\// ) {
			$commentline = $lines[$i];
		}
		elsif ( index( $lines[$i], '<pre>' ) > -1 ) {

			# testable statement
			$testablestatement = "";
			my $index = $i + 1;
			while ( $index < $max && !( $lines[$index] =~ /^\s*\<\/pre/ ) ) {
				$testablestatement .= $lines[$index];
				$index++;
			}
			$i = $index;    # update $i
		}
		elsif ( $lines[$i] =~ /^\s*\{/ ) {

			# expected results table

			my $index = $i + 1;

			my $wikitable = "";
			while ( $index < $max && !( $lines[$index] =~ /^\s*\|\}/ ) ) {

				# gather up table lines excluding {| class='wikitable' line and closing |}
				$wikitable .= $lines[$index];
				$index++;
			}
			$i = $index;

			my %test = (
				TABLE   => $wikitable,
				TEST    => $testablestatement,
				COMMENT => $commentline
			);
			@$tests[$counter] = \%test;
			$counter++;
		}
	}
	return %alltests;
}

#/////////////////////////////////////////////////////////////////////////////

=head2 get_test_elements
 
 Parameters  : testable statement
 Returns     : HTML or SVG elements defining the test  
 Description : gets the html/svg to be placed in the template
 
=cut

sub get_test_elements($) {

	# substring from the first < to the last > before then, then may be missing
	my $start = index( $_[0], '<' );
	my $elements = substr( $_[0], $start );

	my $then = index( $elements, 'then' );
	my $end;
	if ( $then == -1 ) {

		# then is missing
		$end = rindex( $elements, '>' ) + 1;
	}
	else {
		$end = rindex( $elements, '>', $then ) + 1;
	}

	return substr( $elements, 0, $end );
}

#/////////////////////////////////////////////////////////////////////////////

=head2 type_of_test_element
 
 Parameters  : elements
 Returns     : element (type) used in test  
 Description : returns the element type with the id="test"
 
=cut

sub type_of_test_element($) {
	my $start = test_element_open( $_[0] ) + 1;
	my $end = index( $_[0], " ", $start );
	return substr( $_[0], $start, $end - $start );
}

#/////////////////////////////////////////////////////////////////////////////

=head2 test_element_open
 
 Parameters  : svg
 Returns     : position of open angle bracket of the test element  
 Description : finds the postion of the open angle bracket of the test element
 
=cut

sub test_element_open($) {

	# finds id = test, uses double quotes, then single quotes and weakens the test if needed
	# find id="test"
	my $id = index( $_[0], 'id="test"' );
	if ( $id == -1 ) {

		# not found, try single quotes
		$id = index( $_[0], "id='test'" );
		if ( $id == -1 ) {

			# not found, try weaker test
			$id = index( $_[0], '"test"' );
			if ( $id == -1 ) {

				# not found, try single quotes
				$id = index( $_[0], "'test'" );
			}
		}
	}

	#	find the first open angle bracket before the id
	return rindex( $_[0], '<', $id );
}

#/////////////////////////////////////////////////////////////////////////////

=head2 file_to_json
 
 Parameters  : type, file string
 Returns     : json string for object
 Description : converts a type and a file string into a JSON object string.
 
=cut

sub file_to_json($$) {
	my $json = "    { \"type\":\"" . $_[0] . "\", \n    \"test\":\n";
	$json .= "\"" . join( '\"', split( '"', $_[1] ) ) . "\"\n }";
}

#/////////////////////////////////////////////////////////////////////////////

=head2 clean_for_output
 
 Parameters  : string to clean
 Returns     : cleaned string
 Description : cleans a string to be put into a csv file.
 
=cut

sub clean_for_output($) {
	if ( !$_[0] || length( $_[0] ) == 0 ) {
		return "";
	}

	my $string = $_[0];

	# replace comma with a space
	$string =~ s/,+/ /og;

	# remove an equal sign
	$string =~ s/\s*=\s*//og;

	# remove quotes
	$string =~ s/\"*\'*//og;

	# remove a newline
	$string =~ s/\n*//og;

	# remove excess whitespace
	$string =~ s/\s+/ /og;
	$string =~ s/^\s//o;
	$string =~ s/\s$//o;
	return $string;
}

#/////////////////////////////////////////////////////////////////////////////

=head2 clean_quote
 
 Parameters  : string to clean and quote
 Returns     : cleaned quoted string
 Description : cleans and quotes a string for JSON
 
=cut

sub clean_quote($) {
	$string = $_[0];

	#	# replace comma with a space
	#	$string =~ s/,+/ /og;
	#
	#	# remove an equal sign
	#	$string =~ s/\s*=\s*//og;
	#
	# remove quotes
	$string =~ s/\"*\'*//og;

	# remove a newline
	$string =~ s/\n*//og;

	# remove excess whitespace
	$string =~ s/\s+/ /og;
	$string =~ s/^\s//o;
	$string =~ s/\s$//o;

	return '"' . $string . '"';
}

#/////////////////////////////////////////////////////////////////////////////

=head2 guess_section
 
 Parameters  : type, issvg
 Returns     : section URI
 Description : guesses the section of the spec for the test.
 
=cut

sub guess_section($$) {
	my $var = $_[0];
	if ( $var == NAME || $var == DESCRIPTION ) {
		return "https://rawgit.com/w3c/aria/master/svg-aam/svg-aam.html#mapping_additional_nd";
	}

	if ( $var == ROLE_NONE || $var == ROLE_PRESENTATION ) {
		return "https://rawgit.com/w3c/aria/master/svg-aam/svg-aam.html#exclude_elements";
	}

	if ( $var == ROLE ) {

		return ( $_[1] )
		  ? "https://rawgit.com/w3c/aria/master/svg-aam/svg-aam.html#mapping_role_table"
		  : "http://rawgit.com/w3c/aria/master/core-aam/core-aam.html#mapping_role";
	}
	if ( $var == NOT_IN_TREE ) {
		return "https://rawgit.com/w3c/aria/master/svg-aam/svg-aam.html#mapping_role_table";
	}

	return '';
}

#/////////////////////////////////////////////////////////////////////////////

=head2 get_type_enum
 
 Parameters  : type string
 Returns     : type enum
 Description : returns the type enum 
 
=cut

sub get_type_enum($) {
	my $var = lc( $_[0] );
	$var =~ s/^\s+|\s+$//go;    #trim whitespace

	if ( $var eq "not in accessibility tree" ) {
		return NOT_IN_TREE;
	}

	if ( $var eq "name" ) {
		return NAME;
	}

	if ( $var eq "description" ) {
		return DESCRIPTION;
	}

	if ( $var eq "role is none" ) {
		return ROLE_NONE;
	}

	if ( $var eq "role is presentation" ) {
		return ROLE_PRESENTATION;
	}

	if ( index( $var, "role" ) > -1 ) {
		return ROLE;
	}

	# language tests are often more wordy
	if ( index( $var, "name" ) > -1 ) {
		return NAME;
	}

	if ( index( $var, "description" ) > -1 ) {
		return DESCRIPTION;
	}

	return 0;
}

#/////////////////////////////////////////////////////////////////////////////

=head2 add_json_test_object
 
 Parameters  : filehandle, type, element, comment, html, svg, expected result as JSON
 Returns     :  none
 Description : writes a json object for a test to the db file
 
=cut

sub add_json_test_object($$$$$$$) {
	my $fh = shift;

	my $json = "\n {\n  \"testing\":\"";
	$json .= clean_for_output( $_[0] );
	$json .= "\",\n  \"element\":\"";
	$json .= clean_for_output( $_[1] );
	$json .= "\",\n  \"comment\":\"";
	$json .= clean_for_output( $_[2] );
	$json .= "\",\n  \"expected\": ";

	if ( $_[5] ) {    # If expected result is empty, it is printed out in the main program as a bad file
		$json .= $_[5];
	}

	$json .= "  ,\n  \"files\": [\n";
	$json .= file_to_json( "html", $_[3] );

	if ( $_[4] ) {
		$json .= ",\n";
		$json .= file_to_json( "svg", $_[4] );
	}

	$json .= "\n ]}";

	print $fh $json;
}

# main program
# convert wiki markup with tables into JSON for testing
# when switching groups of tests a line will begin and end with ==
# a test starts with a comment describing the test ; //  (whitespace may vary)
# expected result starts with   then  and has a wiki table

# file names
# where the wiki markup is - the wiki markup copied into this file
#my $wiki_markup_filename = 'c:\\bob\\wikitable.txt';
#my $wiki_markup_filename = 'c:\\bob\\wikitableLang.txt';
#my $wiki_markup_filename = 'c:\\bob\\wikitablearia11.txt';
#my $wiki_markup_filename = 'c:\\bob\\searchbox.txt';
#my $wiki_markup_filename = 'c:\\bob\\menuitem.txt';
#my $wiki_markup_filename = 'c:\\bob\\applicationRole.txt';
my $wiki_markup_filename = 'c:\\bob\\aria11roles.txt';
#my $wiki_markup_filename = 'c:\\bob\\svg11.txt';
#my $wiki_markup_filename = 'c:\\bob\\article.txt';

#my $wiki_markup_filename = 'c:\\bob\\test2.txt';

#directories to write test files to and JSON files
$test_dir = "c:\\test\\";

# JSON file
my $json_filename = $test_dir . "aria11test.json";

# create the test directories if they don't exist
mkdir($test_dir) unless ( -d $test_dir );

# alltests is a hash with all the tests from the wiki markup
my %alltests = read_wiki_file($wiki_markup_filename);

# for HTML testing $issvg = 0 and the array size of files will be one, if $issvg > 0 then files size will be 2 and an svg file test will be added
my $issvg = 1;
my $html_template;

# template for HTML test files COMMENT, XLINK and TESTABLE_STATEMENT are replaced
if ($issvg) {
	$html_template = '<!DOCTYPE html>
<html>
   <head>
      <meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
      <title> COMMENT </title>
   </head>
   <body>
   <svg xmlns="http://www.w3.org/2000/svg" XLINK >
     TESTABLE_STATEMENT
   </svg>
   </body>
</html>';
}
else {
	$html_template = '<!DOCTYPE html>
<html>
   <head>
      <meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
      <title> COMMENT </title>
   </head>
   <body>
 	  TESTABLE_STATEMENT
   </body>
</html>';
}

# template for SVG test files XLINK and TESTABLE_STATEMENT are replaced
my $svg_template = ($issvg)
  ? '<svg xmlns="http://www.w3.org/2000/svg" XLINK >
	TESTABLE_STATEMENT
</svg>'
  : "";

#open the json file
open( my $json_html, ">:encoding(UTF-8)", $json_filename )
  || die "can't open UTF-8 encoded $json_filename: $!";

# start the json file (open an array)
print $json_html "[";

my $counter = 0;

# main loop, building test files and JSON files
foreach my $key ( keys %alltests ) {
	my $tests = $alltests{$key};

# simplify the test type want to keep track of - description, name, role, not in accessiblity tree and roles NONE, PRESENTATION
	my $type = $key;
	$type =~ s/calculation//o;
	$type =~ s/Role/role/o;
	my $testType  = get_type_enum($type);
	my $section   = guess_section( $testType, $issvg );
	my $typeCount = 0;

	# make the testfiles and populate the database files
	foreach my $test ( @{ $alltests{$key} } ) {

		my $testsnippet = get_test_elements( $test->{TEST} );
		my $element     = type_of_test_element($testsnippet);
		my $comment     = $test->{COMMENT};
		$comment =~ s/;\s*\/\///;

		# make the HTML test file
		my $h = populate_test_template( $html_template, $comment, $testsnippet );

		# make the SVG test file
		my $s;
		if ($issvg) {
			$s = populate_test_template( $svg_template, $comment, $testsnippet );
		}

		# add the test to the JSON
		if ( $counter > 0 ) {

			# add a comma before adding the new test
			print $json_html ',';
		}

		my $expectedJson = get_json_from_table( $test->{TABLE}, 14 );
		if ( !$expectedJson ) {
			print $comment . " element: " + $element + " failed to create a good test   \n";
		}

		add_json_test_object( $json_html, $type, $element, $comment, $h, $s, $expectedJson );
		$counter++;
		$typeCount++;
	}

	print $typeCount. " tests for " . $type . "\n";

}

# ending the process, closing files
print $counter. " tests were written \n";

# finish and close JSON database files
print $json_html "\n]";
close $json_html;

print 'program exited normally';
